package rice.p2p.past.gc;

import java.math.*;
import java.security.*;
import java.util.*;

import rice.p2p.commonapi.*;
import rice.p2p.util.*;

/**
 * @(#) GCIdSet.java
 *
 * Internal representation of a set of GCIds
 * 
 * @version $Id: GCIdSet.java 2302 2005-03-11 00:58:26Z jeffh $
 *
 * @author Alan Mislove
 * @author Peter Druschel
 */
public class GCIdSet implements IdSet {
  
  // internal representation of the ids
  protected IdSet ids;
  
  // interal list of the timeouts
  protected SortedMap timeouts;
  
  /**
   * Constructor
   */
  protected GCIdSet(IdFactory factory) {
    this.ids = factory.buildIdSet();
    this.timeouts = new RedBlackMap();
  }
  
  /**
   * Constructor
   */
  protected GCIdSet(IdSet set, SortedMap timeouts) {
    this.ids = set;
    this.timeouts = timeouts;
  }
  
  /**
   * return the number of elements
   */
  public int numElements() {
    return ids.numElements();
  }
  
  /**
   * add a member
   * @param id the id to add
   */
  public void addId(Id id) {
    GCId gcid = (GCId) id;
    
    ids.addId(gcid.getId());
    timeouts.put(gcid.getId(), new GCPastMetadata(gcid.getExpiration()));
  }
  
  /**
   * remove a member
   * @param id the id to remove
   */
  public void removeId(Id id) {
    GCId gcid = (GCId) id;
    ids.removeId(gcid.getId());
    timeouts.remove(gcid.getId());
  }
  
  /**
   * test membership
   * @param id the id to test
   * @return true of id is a member, false otherwise
   */
  public boolean isMemberId(Id id) {
    GCId gcid = (GCId) id;
    return ids.isMemberId(gcid.getId());
  }
  
  /**
   * return a subset of this set, consisting of the member ids in a given range
   * @param from the lower end of the range (inclusive)
   * @param to the upper end of the range (exclusive)
   * @return the subset
   */
  public IdSet subSet(IdRange range) {
    return new GCIdSet(ids.subSet(((GCIdRange) range).getRange()), timeouts);
  }
  
  /**
   * return an iterator over the elements of this set
   * @return the interator
   */
  public Iterator getIterator() {
    return new Iterator() {
      Iterator i = ids.getIterator();
      
      public boolean hasNext() { return i.hasNext(); }
      public Object next() { return getGCId((Id) i.next()); }
      public void remove() { throw new UnsupportedOperationException("Remove on GCIdSet()!"); }
    };
  }
  
  protected GCId getGCId(Id id) {
    GCPastMetadata metadata = (GCPastMetadata) timeouts.get(id);
    
    if (metadata != null)
      return new GCId(id, metadata.getExpiration());
    else
      return new GCId(id, GCPastImpl.DEFAULT_EXPIRATION);
  }
  
  /**
   * return this set as an array
   * @return the array
   */
  public Id[] asArray() {
    Id[] array = ids.asArray();
    
    for (int i=0; i<array.length; i++)
      array[i] = getGCId(array[i]);
    
    return array;
  }
  
  /**
   * return a hash of this set
   *
   * @return the hash of this set
   */
  public byte[] hash() {
    throw new UnsupportedOperationException("hash on GCIdSet()!");
  }
  
  /**
   * Determines equality
   *
   * @param other To compare to
   * @return Equals
   */
  public boolean equals(Object o) {
    GCIdSet other = (GCIdSet) o;
    
    if (numElements() != other.numElements())
      return false;
    
    Iterator i = ids.getIterator();
    while (i.hasNext())
      if (! other.isMemberId((Id) i.next()))
        return false;
    
    return true;
  }
  
  /**
   * Returns the hashCode
   *
   * @return hashCode
   */
  public int hashCode() {
    return ids.hashCode();
  }
  
  /**
   * Prints out the string
   *
   * @return A string
   */
  public String toString() {
    return "{GCIdSet of size " + numElements() + "}";
  }
  
  /**
   * Clones this object
   *
   * @return a clone
   */
  public Object clone() {
    return new GCIdSet(ids, timeouts);
  }
  
  /**
    * Returns a new, empty IdSet of this type
   *
   * @return A new IdSet
   */
  public IdSet build() {
    return new GCIdSet(ids.build(), new RedBlackMap());
  }
}
